"""OpenVPN module for AsusRouter."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

_LOGGER = logging.getLogger(__name__)


class AsusOVPNClient(IntEnum):
    """Asus OpenVPN client state."""

    UNKNOWN = -999
    ERROR = -1
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2

    OFF = DISCONNECTED
    ON = CONNECTING


class AsusOVPNServer(IntEnum):
    """Asus OpenVPN server state."""

    UNKNOWN = -999
    ERROR = -1
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2

    OFF = DISCONNECTED
    ON = CONNECTING


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusOVPNClient | AsusOVPNServer,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the OpenVPN state."""

    # Check if arguments are available
    if not arguments:
        arguments = {}

    # Get the id from arguments
    vpn_id = arguments.get("id")
    if not vpn_id:
        _LOGGER.debug("No VPN id found in arguments")
        return False

    # Get the correct service call
    party = "client" if isinstance(state, AsusOVPNClient) else "server"
    match state:
        case AsusOVPNClient.ON:
            service = f"start_vpn{party}{vpn_id}"
        case AsusOVPNClient.OFF:
            service = f"stop_vpn{party}{vpn_id}"
        case _:
            _LOGGER.debug("Unknown state %s", state)
            return False

    # Prepare the arguments
    arguments = {}

    _LOGGER.debug(
        "Triggering state set with parameters: service=%s, arguments=%s",
        service,
        arguments,
    )

    # Run the service
    return await callback(
        service=service,
        arguments=arguments,
        apply=True,
        expect_modify=expect_modify,
    )



"""Command endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.tools.readers import read_json_content


def read(content: str) -> dict[str, Any]:  # pylint: disable=unused-argument
    """Read state data"""

    # Read the json content
    command: dict[str, Any] = read_json_content(content)

    return command



"""Data finder module."""

from __future__ import annotations

from enum import Enum
from typing import Callable, Optional

from asusrouter.modules.attributes import AsusRouterAttribute
from asusrouter.modules.data import AsusData
from asusrouter.modules.endpoint import Endpoint
from asusrouter.modules.parental_control import (
    KEY_PARENTAL_CONTROL_MAC,
    KEY_PARENTAL_CONTROL_NAME,
    KEY_PARENTAL_CONTROL_STATE,
    KEY_PARENTAL_CONTROL_TIMEMAP,
    KEY_PARENTAL_CONTROL_TYPE,
)
from asusrouter.modules.wlan import gwlan_nvram_request, wlan_nvram_request
from asusrouter.tools import converters


class AsusDataMerge(str, Enum):
    """AsusRouter data merge class."""

    ALL = "all"
    ANY = "any"


class AsusDataFinder:
    """AsusRouter data finder class."""

    def __init__(
        self,
        endpoint: list[Endpoint] | Endpoint,
        merge: AsusDataMerge = AsusDataMerge.ANY,
        request: Optional[list[tuple[str, ...]]] = None,
        nvram: Optional[list[str] | str] = None,
        method: Optional[Callable] = None,
        arguments: Optional[AsusRouterAttribute] = None,
    ) -> None:
        """Initialize the data finder."""

        # Set the endpoint as list even if it's a single endpoint
        if isinstance(endpoint, Endpoint):
            endpoint = [endpoint]
        self.endpoint = endpoint

        # Set the merge
        self.merge = merge

        # Set the request and append nvram hooks to the request
        self.request = request or []
        if nvram:
            nvram_request = converters.nvram_get(nvram)
            if nvram_request:
                self.request.extend(nvram_request)

        # Set the method and arguments
        self.method = method
        self.arguments = arguments


# A constant list of requests for fetching data
ASUSDATA_REQUEST = {
    "devices": {
        ("get_clientlist", ""),
    },
    "main": {
        ("cpu_usage", "appobj"),
        ("memory_usage", "appobj"),
        ("netdev", "appobj"),
        ("wanlink_state", "appobj"),
    },
}

ASUSDATA_NVRAM = {
    "light": ["led_val"],
    "parental_control": [
        KEY_PARENTAL_CONTROL_MAC,
        KEY_PARENTAL_CONTROL_NAME,
        KEY_PARENTAL_CONTROL_STATE,
        KEY_PARENTAL_CONTROL_TIMEMAP,
        KEY_PARENTAL_CONTROL_TYPE,
    ],
    "port_forwarding": [
        "vts_rulelist",
        "vts_enable_x",
    ],
}

ASUSDATA_ENDPOINT_APPEND = {
    Endpoint.PORT_STATUS: {
        "node_mac": AsusRouterAttribute.MAC,
    }
}


# A map of endptoins to get data from
ASUSDATA_MAP = {
    AsusData.AIMESH: AsusDataFinder(Endpoint.ONBOARDING),
    AsusData.BOOTTIME: AsusDataFinder(Endpoint.DEVICEMAP),
    AsusData.CLIENTS: AsusDataFinder(
        [Endpoint.ONBOARDING, Endpoint.UPDATE_CLIENTS], AsusDataMerge.ALL
    ),
    AsusData.CPU: AsusDataFinder(Endpoint.HOOK, request=ASUSDATA_REQUEST["main"]),
    AsusData.DEVICEMAP: AsusDataFinder(Endpoint.DEVICEMAP),
    AsusData.FIRMWARE: AsusDataFinder(Endpoint.FIRMWARE),
    AsusData.GWLAN: AsusDataFinder(
        Endpoint.HOOK,
        method=gwlan_nvram_request,
        arguments=AsusRouterAttribute.WLAN_LIST,
    ),
    AsusData.LED: AsusDataFinder(Endpoint.HOOK, nvram=ASUSDATA_NVRAM["light"]),
    AsusData.NETWORK: AsusDataFinder(Endpoint.HOOK, request=ASUSDATA_REQUEST["main"]),
    AsusData.NODE_INFO: AsusDataFinder(Endpoint.PORT_STATUS),
    AsusData.OPENVPN: AsusDataFinder(
        [Endpoint.VPN, Endpoint.DEVICEMAP], AsusDataMerge.ANY
    ),
    AsusData.PARENTAL_CONTROL: AsusDataFinder(
        Endpoint.HOOK, nvram=ASUSDATA_NVRAM["parental_control"]
    ),
    AsusData.PORT_FORWARDING: AsusDataFinder(
        Endpoint.HOOK, nvram=ASUSDATA_NVRAM["port_forwarding"]
    ),
    AsusData.PORTS: AsusDataFinder([Endpoint.PORT_STATUS, Endpoint.ETHERNET_PORTS]),
    AsusData.RAM: AsusDataFinder(Endpoint.HOOK, request=ASUSDATA_REQUEST["main"]),
    AsusData.SYSINFO: AsusDataFinder(Endpoint.SYSINFO),
    AsusData.TEMPERATURE: AsusDataFinder(Endpoint.TEMPERATURE),
    AsusData.WAN: AsusDataFinder(Endpoint.HOOK, request=ASUSDATA_REQUEST["main"]),
    AsusData.WLAN: AsusDataFinder(
        Endpoint.HOOK,
        method=wlan_nvram_request,
        arguments=AsusRouterAttribute.WLAN_LIST,
    ),
}



"""System module for AsusRouter."""

from __future__ import annotations

from enum import Enum
from typing import Any, Awaitable, Callable, Optional


class AsusSystem(str, Enum):
    """Asus system enum."""

    REBOOT = "reboot"
    RESTART_FIREWALL = "restart_firewall"
    RESTART_HTTPD = "restart_httpd"
    RESTART_WIRELESS = "restart_wireless"


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusSystem,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the LED state."""

    # Check if arguments are available
    if not arguments:
        arguments = {}

    # Run the service
    return await callback(
        service=state.value,
        arguments={},
        apply=True,
        expect_modify=expect_modify,
    )



"""AsusRouter module.

This module contains the main class for interacting with an Asus device."""

from __future__ import annotations

import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Optional

import aiohttp

from asusrouter.connection import Connection
from asusrouter.const import DEFAULT_CACHE_TIME, DEFAULT_TIMEOUT
from asusrouter.error import (
    AsusRouter404Error,
    AsusRouterAccessError,
    AsusRouterConnectionError,
    AsusRouterDataError,
)
from asusrouter.modules.attributes import AsusRouterAttribute
from asusrouter.modules.client import process_client
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.data_finder import (
    ASUSDATA_ENDPOINT_APPEND,
    ASUSDATA_MAP,
    AsusDataFinder,
    AsusDataMerge,
)
from asusrouter.modules.data_transform import transform_clients, transform_network
from asusrouter.modules.endpoint import Endpoint, EndpointControl, process, read
from asusrouter.modules.endpoint.error import AccessError
from asusrouter.modules.flags import Flag
from asusrouter.modules.identity import AsusDevice, collect_identity
from asusrouter.modules.parental_control import ParentalControlRule
from asusrouter.modules.port_forwarding import PortForwardingRule
from asusrouter.modules.service import async_call_service
from asusrouter.modules.state import (
    AsusState,
    get_datatype,
    keep_state,
    save_state,
    set_state,
)
from asusrouter.tools import legacy
from asusrouter.tools.converters import safe_list
from asusrouter.tools.readers import merge_dicts, readable_mac

_LOGGER = logging.getLogger(__name__)


class AsusRouter:
    """The interface class."""

    _cache_time: float = DEFAULT_CACHE_TIME
    _hostname: str

    _identity: Optional[AsusDevice] = None

    _state: dict[AsusData, AsusDataState] = {}

    _flags: Flag = Flag()

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        port: Optional[int] = None,
        use_ssl: bool = False,
        cache_time: Optional[float] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):  # pylint: disable=too-many-arguments
        """Initialize the interface."""

        _LOGGER.debug("Initializing a new interface to `%s`", hostname)

        # Set the cache time
        if cache_time:
            self._cache_time = cache_time

        # Set the host
        self._hostname = hostname

        # Create a connection
        self._connection = Connection(
            hostname=hostname,
            username=username,
            password=password,
            port=port,
            use_ssl=use_ssl,
            session=session,
        )

    # ---------------------------
    # Connection-related methods -->
    # ---------------------------

    async def async_connect(self) -> bool:
        """Connect to the device and get its identity."""

        _LOGGER.debug("Triggered method async_connect")

        # Connect to the device
        try:
            result = await self._connection.async_connect()
            if result is False:
                return False
        except Exception as ex:  # pylint: disable=broad-except
            raise ex

        # Get the device identity
        return await self.async_get_identity() is not None

    async def async_disconnect(self) -> bool:
        """Disconnect from the device."""

        _LOGGER.debug("Triggered method async_disconnect")

        # Disconnect from the device
        try:
            await self._connection.async_disconnect()
        except Exception as ex:  # pylint: disable=broad-except
            await self._async_handle_exception(ex)

        return True

    async def _async_drop_connection(self) -> None:
        """Drop the connection in case we know it cannot reply
        due to our last actions."""

        _LOGGER.debug("Triggered method _async_drop_connection")

        self._connection.reset_connection()

    async def _async_handle_exception(self, ex: Exception) -> None:
        """Handle exceptions."""

        _LOGGER.debug("Triggered method _async_handle_exception")

        raise ex

    async def _async_handle_reboot(self) -> None:
        """Handle reboot."""

        _LOGGER.debug("Triggered method _async_handle_reboot")

        led_state = self._state.get(AsusData.LED)
        if led_state and led_state.data:
            _LOGGER.debug("Restoring LED state")
            await keep_state(
                self.async_run_service, led_state.data["state"], self._identity
            )

    # ---------------------------
    # <-- Connection-related methods
    # ---------------------------

    # ---------------------------
    # Identity-related methods -->
    # ---------------------------

    async def async_get_identity(self, force: bool = False) -> AsusDevice:
        """Get the device identity."""

        _LOGGER.debug("Triggered method async_get_identity")

        # Check whether we already have the identity and not forcing a refresh
        if self._identity and not force:
            return self._identity

        # Collect the identity
        self._identity = await collect_identity(
            api_hook=self.async_api_hook,
            api_query=self.async_api_query,
        )

        # Return new identity
        return self._identity

    # ---------------------------
    # <-- Identity-related methods
    # ---------------------------

    # ---------------------------
    # Request-related methods -->
    # ---------------------------

    def _get_attribute(self, attribute: Optional[AsusRouterAttribute]) -> Optional[Any]:
        """Get an attribute value."""

        if not attribute:
            return None

        match (attribute):
            case AsusRouterAttribute.MAC:
                if self._identity:
                    return self._identity.mac
            case AsusRouterAttribute.WLAN_LIST:
                if self._identity:
                    return self._identity.wlan

        return None

    async def _check_flags(self) -> None:
        """Check flags."""

        _LOGGER.debug("Triggered method _check_flags")

        flags: dict[str, bool] = {}

        state_flags = self._state.get(AsusData.FLAGS)
        if isinstance(state_flags, AsusDataState) and isinstance(
            state_flags.data, dict
        ):
            flags = state_flags.data

        print("FLAGS: ", flags)

        if flags.get("reboot", False) is True:
            _LOGGER.debug("Reboot flag is set")
            await self._async_handle_reboot()

    async def async_api_query(
        self, endpoint: Endpoint | EndpointControl, payload: Optional[str] = None
    ):
        """A wrapper for the connection query method."""

        if endpoint in ASUSDATA_ENDPOINT_APPEND:
            payload = payload or ""
            for key, attribute in ASUSDATA_ENDPOINT_APPEND[endpoint].items():
                value = self._get_attribute(attribute)
                if value:
                    payload += f"{key}={value};"
            # Remove trailing semicolon
            payload = payload[:-1]

        _LOGGER.debug("Triggered method async_api_query: %s | %s", endpoint, payload)

        return await self._connection.async_query(endpoint, payload)

    async def async_api_load(
        self,
        endpoint: Endpoint | EndpointControl,
        request: str = "",
        retry: int = 0,
    ) -> dict[str, Any]:
        """Load API endpoint with optional request."""

        _LOGGER.debug("Triggered method async_api_load: %s", endpoint)

        # Load the endpoint
        try:
            status, _, content = await self.async_api_query(endpoint, request)
        except AsusRouter404Error:
            _LOGGER.debug("Endpoint %s not found", endpoint)
            return {}
        except AsusRouterAccessError as ex:
            # Check whether we are not connected
            args = ex.args
            if args[1] == AccessError.AUTHORIZATION:
                # Mark the connection as dropped
                await self._async_drop_connection()
                # Wait before repeating the request
                await asyncio.sleep(1 + retry * 3)
                # Repeat request once more and see what happens
                return await self.async_api_load(endpoint, request, True)
            # Otherwise just raise the exception
            raise ex

        # Log status
        _LOGGER.debug("Response %s received from %s", status, endpoint)

        # Try to read the content
        try:
            result = read(endpoint, content)
        except json.JSONDecodeError as ex:
            # Not like this is supposed to happen, but just in case
            _LOGGER.debug("Failed to read content from %s", endpoint)
            _LOGGER.debug("Content: %s", content)
            # Just repeat request once more and see what happens
            # Only if we haven't tried already
            if not retry:
                return await self.async_api_load(endpoint, request, True)
            raise AsusRouterDataError(
                "Something went wrong while reading the content"
            ) from ex

        # Check if we need to drop the connection
        run_service = result.get("run_service", None)
        if run_service in ("restart_httpd", "reboot"):
            await self._async_drop_connection()

        return result

    async def async_api_hook(self, request: str) -> dict[str, Any]:
        """Hook to the device API. Hooks are used to fetch data from the device."""

        _LOGGER.debug("Triggered method async_api_hook: %s", request)

        return await self.async_api_load(
            endpoint=Endpoint.HOOK,
            request=f"hook={request}",
        )

    async def async_api_command(
        self,
        commands: Optional[dict[str, str]],
        endpoint: Endpoint | EndpointControl = EndpointControl.COMMAND,
    ):
        """Send a command to the device."""

        _LOGGER.debug("Triggered method async_api_command: %s | %s", endpoint, commands)

        return await self.async_api_load(
            endpoint=endpoint,
            request=str(commands),
        )

    # ---------------------------
    # <-- Request-related methods
    # ---------------------------

    def _where_to_get_data(self, datatype: AsusData) -> Optional[AsusDataFinder]:
        """Get the list of endpoints to get data from."""

        _LOGGER.debug("Triggered method _where_to_get_data")

        # Check that device identity is available
        if not self._identity:
            _LOGGER.debug("No device identity available")
            return None

        # Get the map
        data_map = ASUSDATA_MAP.get(datatype)
        if not data_map:
            return None

        # Check if endpoints are available
        for endpoint in data_map.endpoint:
            # Check endpoint availability in identity
            if self._identity.endpoints and self._identity.endpoints.get(endpoint) in (
                False,
                None,
            ):
                # Remove the endpoint from the map
                data_map.endpoint.remove(endpoint)

        _LOGGER.debug("Endpoints to check: %s", data_map.endpoint)

        return data_map

    def _transform_data(self, datatype: AsusData, data: Any) -> Any:
        """Transform data if needed."""

        _LOGGER.debug("Triggered method _transform_data for `%s`", datatype)

        if datatype == AsusData.CLIENTS:
            _LOGGER.debug("Transforming clients data")
            return transform_clients(data, self._state.get(AsusData.CLIENTS))

        if datatype == AsusData.NETWORK:
            _LOGGER.debug("Transforming network data")
            return transform_network(
                data,
                self._identity.services if self._identity else [],
                self._state.get(AsusData.NETWORK),
            )

        return data

    def _check_state(self, datatype: Optional[AsusData]) -> None:
        """Make sure the state object is available."""

        _LOGGER.debug("Triggered method _check_state")

        if not datatype:
            return

        # Add state object but make sure it's marked expired
        if datatype not in self._state:
            self._state[datatype] = AsusDataState(
                timestamp=(
                    datetime.now(timezone.utc) - timedelta(seconds=2 * self._cache_time)
                )
            )

    async def async_get_data(self, datatype: AsusData, force: bool = False) -> Any:
        """Generic method to get data from the device."""

        # Check if we have a state object for this data
        self._check_state(datatype)

        # If state object is active, wait for it to finish and return the data
        if self._state[datatype].active:
            try:
                _LOGGER.debug("Already in progress. Waiting for data to be fetched")
                await asyncio.wait_for(
                    self._state[datatype].inactive_event.wait(), DEFAULT_TIMEOUT
                )
            except asyncio.TimeoutError:
                _LOGGER.debug("Timeout while waiting for data. Will try fetching again")

        # Check if we have the data already and not forcing a refresh
        if self._state[datatype].data and not force:
            # Check if the data is younger than the cache time
            if datetime.now(timezone.utc) - self._state[datatype].timestamp < timedelta(
                seconds=self._cache_time
            ):
                _LOGGER.debug(
                    "Using cached data for `%s`: %s",
                    datatype,
                    self._state[datatype].data,
                )
                # Return the cached data
                return self._state[datatype].data
            _LOGGER.debug("Data for %s is too old. Fetching", datatype)

        # Mark the data as active
        self._state[datatype].start()

        # Get the data finder
        data_finder = self._where_to_get_data(datatype)

        # Check if we have a data finder
        if not data_finder:
            _LOGGER.debug("No data finder for %s", datatype)
            return {}

        # The data we are looking for
        data = {}
        result: dict[AsusData, Any] = {}

        try:
            for endpoint in data_finder.endpoint:
                # Get the data from the endpoint
                request = "hook=" if endpoint == Endpoint.HOOK else ""
                for item in data_finder.request:
                    key, value = item
                    request += f"{key}({value});"
                if data_finder.method:
                    argument = self._get_attribute(data_finder.arguments)
                    request += (
                        data_finder.method(argument)
                        if argument
                        else data_finder.method()
                    )
                # Check that we are not fetching this data already

                # Fetch the data
                data = await self.async_api_load(endpoint, request)

                # Make sure, identity is available
                if not self._identity:
                    self._identity = await self.async_get_identity()

                result = merge_dicts(
                    result,
                    process(
                        endpoint,
                        data,
                        self._state,
                        self._identity.firmware,
                        self._identity.wlan,
                    ),
                )

                # Check if we have data and data finder merge is ANY
                if result and data_finder.merge == AsusDataMerge.ANY:
                    break

            # Save the data state
            for key, value in result.items():
                # Transform data if needed
                value = self._transform_data(key, value)
                # Save the data
                result[key] = value
                # Update the state
                if key not in self._state:
                    self._state[key] = AsusDataState()
                self._state[key].update(value)
        except (AsusRouterConnectionError, AsusRouterDataError):
            return self._state[datatype].data

        # Check flags
        await self._check_flags()

        # Return the data we were looking for
        _LOGGER.debug(
            "Returning data for `%s` with object type `%s`",
            datatype,
            type(self._state[datatype].data),
        )
        return self._state[datatype].data

    # ---------------------------
    # Service-related methods -->
    # ---------------------------

    async def async_run_service(
        self,
        service: str,
        arguments: Optional[dict[str, Any]] = None,
        apply: bool = False,
        expect_modify: bool = True,
        drop_connection: bool = False,
    ) -> bool:
        """Run a service."""

        _LOGGER.debug("Triggered method async_run_service")

        # Run the service
        result = await async_call_service(
            self.async_api_command,
            service,
            arguments,
            apply,
            expect_modify,
        )

        if drop_connection:
            await self._async_drop_connection()

        return result

    async def async_set_state(
        self,
        state: AsusState,
        arguments: Optional[dict[str, Any]] = None,
        expect_modify: bool = False,
    ) -> bool:
        """Set the state."""

        _LOGGER.debug("Triggered method async_set_state")

        _LOGGER.debug(
            "Setting state `%s` with arguments `%s`. Expecting modify: `%s`",
            state,
            arguments,
            expect_modify,
        )

        result = await set_state(
            self.async_run_service, state, arguments, expect_modify
        )

        if result is True:
            # Check if we have a state object for this data
            self._check_state(get_datatype(state))
            # Save the state
            save_state(state, self._state)

        return result

    # ---------------------------
    # <-- Service-related methods
    # ---------------------------

    # ---------------------------
    # Legacy methods -->
    # ---------------------------
    # Not tested, not used, not documented
    # This part can be changed / removed in the future
    # It might also not be working properly

    async def async_apply_parental_control_rules(
        self,
        rules: dict[str, ParentalControlRule],
    ) -> bool:
        """Apply parental control rules."""

        request = legacy.compile_parental_control(rules)

        if request:
            return await self.async_run_service(
                service="restart_firewall",
                arguments=request,
                apply=True,
            )

        return False

    async def async_remove_parental_control_rules(
        self,
        macs: Optional[str | list[str]] = None,
        rules: Optional[ParentalControlRule | list[ParentalControlRule]] = None,
        apply: bool = True,
    ) -> dict[str, ParentalControlRule]:
        """Remove parental control rules"""

        _macs = set() if macs is None else set(safe_list(macs))
        rules = [] if rules is None else safe_list(rules)

        # Get current rules
        current_rules: dict = (await self.async_get_data(AsusData.PARENTAL_CONTROL))[
            "rules"
        ]

        # Remove old rules for these MACs
        for mac in _macs:
            current_rules.pop(mac, None)
        for rule in rules:
            current_rules.pop(rule.mac, None)

        # Apply new rules
        if apply:
            await self.async_apply_parental_control_rules(current_rules)

        # Return the new rules
        return current_rules

    async def async_set_parental_control_rules(
        self,
        rules: ParentalControlRule | list[ParentalControlRule],
    ) -> bool:
        """Set parental control rules"""

        rules = safe_list(rules)

        # Remove old rules for these MACs and get the rest of the list
        current_rules = await self.async_remove_parental_control_rules(
            rules, apply=False
        )

        # Add new rules
        for rule in rules:
            current_rules[rule.mac] = rule

        # Apply new rules
        return await self.async_apply_parental_control_rules(current_rules)

    async def async_apply_port_forwarding_rules(
        self,
        rules: list[PortForwardingRule],
    ) -> bool:
        """Apply port forwarding rules"""

        request = legacy.compile_port_forwarding(rules)

        if request:
            return await self.async_run_service(
                service="restart_firewall",
                arguments=request,
                apply=True,
            )

        return False

    async def async_remove_port_forwarding_rules(
        self,
        ips: str | list[str] | None = None,
        rules: PortForwardingRule | list[PortForwardingRule] | None = None,
        apply: bool = True,
    ) -> list[PortForwardingRule]:
        """Remove port forwarding rules"""

        ips = set() if ips is None else set(safe_list(ips))
        rules = [] if rules is None else safe_list(rules)

        # Get current rules
        current_rules: list[PortForwardingRule] = (
            await self.async_get_data(AsusData.PORT_FORWARDING)
        )["rules"]

        # Remove all rules for these IPs
        current_rules = [rule for rule in current_rules if rule.ip_address not in ips]
        # Remove exact rules
        for rule_to_find in rules:
            current_rules = [
                rule
                for rule in current_rules
                if not (
                    rule.ip_address == rule_to_find.ip_address
                    and rule.port_external == rule_to_find.port_external
                    and rule.protocol == rule_to_find.protocol
                    and (
                        rule_to_find.ip_external is None
                        or rule.ip_external == rule_to_find.ip_external
                    )
                    and (rule_to_find.port is None or rule.port == rule_to_find.port)
                )
            ]

        # Apply new rules
        if apply:
            await self.async_apply_port_forwarding_rules(current_rules)

        # Return the new rules
        return current_rules

    async def async_set_port_forwarding_rules(
        self,
        rules: PortForwardingRule | list[PortForwardingRule],
    ) -> bool:
        """Set port forwarding rules"""

        rules = safe_list(rules)

        # Get current rules
        current_rules: list[PortForwardingRule] = (
            await self.async_get_data(AsusData.PORT_FORWARDING)
        )["rules"]

        # Add new rules
        for rule in rules:
            current_rules.append(rule)

        # Apply new rules
        return await self.async_apply_port_forwarding_rules(current_rules)

    # ---------------------------
    # <-- Legacy methods
    # ---------------------------

    # ---------------------------
    # Properties -->
    # ---------------------------

    @property
    def connected(self) -> bool:
        """Return connection status."""

        return self._connection.connected

    # ---------------------------
    # <-- Properties
    # ---------------------------

    # ---------------------------
    # General management -->
    # ---------------------------

    async def async_cleanup(self) -> None:
        """Cleanup the connection."""

        self._connection.reset_connection()

    # ---------------------------
    # <-- General management
    # ---------------------------



"""Hook endpoint module."""

from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import Any, Optional

from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.endpoint import data_get
from asusrouter.modules.led import AsusLED
from asusrouter.modules.parental_control import (
    KEY_PARENTAL_CONTROL_MAC,
    KEY_PARENTAL_CONTROL_STATE,
    KEY_PARENTAL_CONTROL_TYPE,
    MAP_PARENTAL_CONTROL_ITEM,
    MAP_PARENTAL_CONTROL_TYPE,
    AsusParentalControl,
    ParentalControlRule,
)
from asusrouter.modules.port_forwarding import (
    KEY_PORT_FORWARDING_LIST,
    KEY_PORT_FORWARDING_STATE,
    AsusPortForwarding,
    PortForwardingRule,
)
from asusrouter.modules.wlan import MAP_GWLAN, MAP_WLAN, Wlan
from asusrouter.tools.converters import (
    safe_int,
    safe_return,
    safe_speed,
    safe_unpack_key,
    safe_unpack_keys,
    safe_usage,
    safe_usage_historic,
)
from asusrouter.tools.readers import read_json_content

from .hook_const import MAP_NETWORK, MAP_WAN

REQUIRE_HISTORY = True
REQUIRE_WLAN = True


def read(content: str) -> dict[str, Any]:
    """Read hook data"""

    hook: dict[str, Any] = read_json_content(content)

    return hook


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process hook data."""

    # For this endpoint, the received data always depends on the sent request.
    # So, we need to check which data is available and process it accordingly.
    # Otherwise, we can accidentally overwrite the data with empty values.

    state: dict[AsusData, Any] = {}

    # Get the passed awrguments
    history: dict[AsusData, AsusDataState] = data_get(data, "history") or {}
    wlan = data_get(data, "wlan") or []

    # CPU
    if "cpu_usage" in data:
        cpu_usage = data.get("cpu_usage", {})
        prev_cpu: Optional[AsusDataState] = history.get(AsusData.CPU)
        state[AsusData.CPU] = (
            process_cpu(cpu_usage, prev_cpu)
            if cpu_usage
            else prev_cpu.data
            if isinstance(prev_cpu, AsusDataState)
            else {}
        )

    # GWLAN
    if (
        "wl0.1_wpa_psk" in data
        or "wl1.1_wpa_psk" in data
        or "wl2.1_wpa_psk" in data
        or "wl3.1_wpa_psk" in data
    ):
        state[AsusData.GWLAN] = process_gwlan(data, wlan)

    # LED
    if "led_val" in data:
        state[AsusData.LED] = {
            "state": AsusLED(safe_int(data.get("led_val"), default=-999))
        }

    # Network
    if "netdev" in data:
        prev_network: Optional[AsusDataState] = history.get(AsusData.NETWORK)
        netdev = data.get("netdev")
        state[AsusData.NETWORK] = (
            process_network(netdev, prev_network)
            if netdev
            else prev_network.data
            if isinstance(prev_network, AsusDataState)
            else {}
        )

    # Parental control
    if KEY_PARENTAL_CONTROL_STATE in data:
        state[AsusData.PARENTAL_CONTROL] = process_parental_control(data)

    # Port forwarding
    if KEY_PORT_FORWARDING_STATE in data:
        state[AsusData.PORT_FORWARDING] = process_port_forwarding(data)

    # RAM
    if "memory_usage" in data:
        memory_usage = data.get("memory_usage", {})
        state[AsusData.RAM] = process_ram(memory_usage) if memory_usage else {}

    # WAN
    if "wanlink_state" in data:
        wanlink_state = data.get("wanlink_state", {})
        state[AsusData.WAN] = process_wan(wanlink_state) if wanlink_state else {}

    # WLAN
    if (
        "wl0_wpa_psk" in data
        or "wl1_wpa_psk" in data
        or "wl2_wpa_psk" in data
        or "wl3_wpa_psk" in data
    ):
        state[AsusData.WLAN] = process_wlan(data, wlan)

    return state


def process_cpu(
    cpu_usage: dict[str, Any], history: Optional[AsusDataState]
) -> dict[str | int, Any]:
    """Process CPU data."""

    cpu = process_cpu_usage(cpu_usage)

    # Get the previous data
    prev_cpu = history.data if history else None

    # Safe calculate of actual usage if previous data is available
    if prev_cpu:
        for item, after in cpu.items():
            if item in prev_cpu:
                before = prev_cpu[item]
                after["usage"] = safe_usage_historic(
                    after["used"], after["total"], before["used"], before["total"]
                )

    return cpu


def process_cpu_usage(raw: dict[str, Any]) -> dict[str | int, Any]:
    """Process CPU usage."""

    # Populate total
    cpu: dict[str | int, Any] = {"total": {"total": 0.0, "used": 0.0}}

    # Process each core
    core = 1
    while f"cpu{core}_total" in raw:
        cpu[core] = {
            "total": int(raw[f"cpu{core}_total"]),
            "used": int(raw[f"cpu{core}_usage"]),
        }
        # Update the total
        cpu["total"]["total"] += cpu[core]["total"]
        cpu["total"]["used"] += cpu[core]["used"]

        core += 1

    return cpu


def process_gwlan(data: dict[str, Any], wlan_list: list[Wlan]) -> dict[str, Any]:
    """Process GWLAN data."""

    gwlan = {}

    for interface in wlan_list:
        index = list(Wlan).index(interface)
        for gid in range(1, 4):
            info = {}
            for pair in MAP_GWLAN:
                key, method = safe_unpack_key(pair)
                info[key.format(f"{index}.{gid}")[6:]] = (
                    method(data.get(key.format(f"{index}.{gid}")))
                    if method
                    else data.get(key.format(f"{index}.{gid}"))
                )
            gwlan[f"{interface.value}_{gid}"] = info

    return gwlan


def process_network(
    netdev: dict[str, Any], history: Optional[AsusDataState]
) -> dict[str, Any]:
    """Process network data."""

    # Calculate RX and TX from the HEX values.
    network = process_network_usage(netdev)

    # Calculate speeds if previous data is available
    if history and history.data:
        time_delta = (datetime.now(timezone.utc) - history.timestamp).total_seconds()
        network = process_network_speed(network, history.data, time_delta)

    return network


def process_network_speed(
    network: dict[str, dict[str, float]],
    prev_network: dict[str, dict[str, float]],
    time_delta: float | None,
) -> dict[str, dict[str, float]]:
    """Calculate network speed for a set period of time"""

    # Check values one by one
    for interface in network:
        # Skip if there is no previous data
        if interface not in prev_network:
            continue

        # Dictionary with speed values
        interface_speed = {}

        # Calculate speed for each traffic type
        for traffic_type, traffic_value in network[interface].items():
            prev_traffic_value = prev_network[interface].get(traffic_type)
            # Calculate speed only if previous value is available
            if prev_traffic_value:
                interface_speed[f"{traffic_type}_speed"] = 8 * safe_speed(
                    traffic_value,
                    prev_traffic_value,
                    time_delta,
                )
                continue
            # Otherwise, set speed to 0
            interface_speed[f"{traffic_type}_speed"] = 0.0

        # Update interface with speed values
        network[interface].update(interface_speed)

    return network


def process_network_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """Process network usage data"""

    network = {}
    for key, key_to_use in MAP_NETWORK.items():
        data = {}
        for traffic_type in ("rx", "tx"):
            # Convert string with HEX value to int
            value = safe_int(raw.get(f"{key}_{traffic_type}"), base=16)
            # Check that value is integer
            if isinstance(value, int):
                data[traffic_type] = value
        if len(data) > 0:
            network[key_to_use] = data

    return network


def process_parental_control(data: dict[str, Any]) -> dict[str, Any]:
    """Process parental control data"""

    parental_control: dict[str, Any] = {}

    # State
    parental_control["state"] = AsusParentalControl(
        safe_int(data.get(KEY_PARENTAL_CONTROL_STATE), default=-999)
    )

    # Rules
    rules = {}

    if data.get(KEY_PARENTAL_CONTROL_MAC) != data.get(KEY_PARENTAL_CONTROL_TYPE):
        as_is = {
            key_to_use: [
                method(temp_element) for temp_element in data[key].split("&#62")
            ]
            for key, key_to_use, method in MAP_PARENTAL_CONTROL_ITEM
        }

        number = len(as_is["mac"])
        for i in range(number):
            rules[as_is["mac"][i]] = ParentalControlRule(
                mac=as_is["mac"][i],
                name=as_is["name"][i],
                type=MAP_PARENTAL_CONTROL_TYPE.get(as_is["type"][i], "unknown"),
                timemap=as_is["timemap"][i] if i < len(as_is["timemap"]) else None,
            )

    parental_control["rules"] = rules.copy()

    return parental_control


def process_port_forwarding(data: dict[str, Any]) -> dict[str, Any]:
    """Process port forwarding data."""

    port_forwarding = {}

    # State
    port_forwarding["state"] = AsusPortForwarding(
        safe_int(data.get(KEY_PORT_FORWARDING_STATE), default=-999)
    )

    # Rules
    pf_list = data.get(KEY_PORT_FORWARDING_LIST)
    if pf_list:
        rules = []
        rule_list = pf_list.split("&#60")
        for rule in rule_list:
            if rule == str():
                continue
            part = rule.split("&#62")
            rules.append(
                PortForwardingRule(
                    name=safe_return(part[0]),
                    ip_address=part[2],
                    port=safe_return(part[3]),
                    protocol=part[4],
                    ip_external=safe_return(part[5]),
                    port_external=part[1],
                )
            )
        port_forwarding["rules"] = rules.copy()

    return port_forwarding


def process_ram(memory_usage: dict[str, Any]) -> dict[str, Any]:
    """Process RAM data."""

    ram: dict[str, Any] = {}
    # Data is in KiB. To get MB as they are shown in the device Web-GUI,
    # should be devided by 1024 (yes, those will be MiB)

    # Populate RAM with known values
    ram = {
        "free": int(memory_usage["mem_free"]),
        "total": int(memory_usage["mem_total"]),
        "used": int(memory_usage["mem_used"]),
    }
    # Calculate usage in percents
    if "used" in ram and "total" in ram:
        ram["usage"] = safe_usage(ram["used"], ram["total"])

    return ram


def process_wan(wanlink_state: dict[str, Any]) -> dict[str, Any]:
    """Process WAN data."""

    wan = {}

    for keys in MAP_WAN:
        key, key_to_use, method = safe_unpack_keys(keys)
        state_value = wanlink_state.get(key)
        if state_value:
            wan[key_to_use] = method(state_value) if method else state_value

    return wan


def process_wlan(data: dict[str, Any], wlan_list: list[Wlan]) -> dict[str, Any]:
    """Process WLAN data."""

    wlan = {}

    for interface in wlan_list:
        index = list(Wlan).index(interface)
        info = {}
        for pair in MAP_WLAN:
            key, method = safe_unpack_key(pair)
            info[key.format(index)[4:]] = (
                method(data.get(key.format(index)))
                if method
                else data.get(key.format(index))
            )
        wlan[interface.value] = info

    return wlan



"""LED module."""

from __future__ import annotations

from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.endpoint import Endpoint
from asusrouter.modules.identity import AsusDevice


class AsusLED(IntEnum):
    """Asus LED state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusLED,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the LED state."""

    # Prepare the arguments
    arguments = {"led_val": state.value}

    # Run the service
    return await callback(
        service="start_ctrl_led",
        arguments=arguments,
        apply=True,
        expect_modify=expect_modify,
    )


async def keep_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusLED = AsusLED.ON,
    identity: Optional[AsusDevice] = None,
) -> bool:
    """Keep the LED state."""

    # Check if identity is available and if endpoints are defined
    if identity is None or not identity.endpoints:
        return False

    # Get the sysinfo
    sysinfo = identity.endpoints.get(Endpoint.SYSINFO)

    # Only when LEDs are off and if sysinfo is available
    if not sysinfo or state == AsusLED.ON:
        return False

    # Toggle the LED
    await set_state(callback, AsusLED.ON)
    await set_state(callback, AsusLED.OFF)

    return True



"""Client module for AsusRouter."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional

from asusrouter.modules.connection import (
    ConnectionState,
    ConnectionType,
    InternetMode,
    get_connection_type,
    get_internet_mode,
)
from asusrouter.modules.ip_address import IPAddressType, read_ip_address_type
from asusrouter.modules.vendor import replace_vendor
from asusrouter.tools.converters import (
    safe_bool,
    safe_float,
    safe_int,
    safe_time_from_delta,
    safe_unpack_key,
)

# # Map of the parameters for connected device
# # value_to_find: [where_to_search, method to convert]
# # List is sorted by importance with the most important first
# MAP_CONNECTED_DEVICE: dict[str, list[SearchKey]] = {
#     "connected_since": [
#         SearchKey("wlConnectTime", time_from_delta),
#     ],
#     CONNECTION_TYPE: [
#         SearchKey(CONNECTION_TYPE),
#         SearchKey("isWL", int_from_str),
#     ],
#     GUEST: [
#         SearchKey(GUEST),
#         SearchKey("isGN", int_from_str),
#     ],
#     "internet_mode": [
#         SearchKey("internetMode"),
#     ],
#     "internet_state": [
#         SearchKey("internetState", bool_from_any),
#     ],
#     IP: [
#         SearchKey(IP),
#     ],
#     "ip_method": [
#         SearchKey("ipMethod"),
#     ],
#     MAC: [
#         SearchKey(MAC),
#     ],
#     NAME: [
#         SearchKey("nickName"),
#         SearchKey(NAME),
#         SearchKey(MAC),
#     ],
#     NODE: [
#         SearchKey(NODE),
#     ],
#     ONLINE: [
#         SearchKey(ONLINE),
#         SearchKey("isOnline", bool_from_any),
#     ],
#     RSSI: [
#         SearchKey(RSSI, int_from_str),
#     ],
#     "rx_speed": [
#         SearchKey("curRx", float_from_str),
#     ],
#     "tx_speed": [
#         SearchKey("curTx", float_from_str),
#     ],
# }


# List of available attributes for a device:
#
# "amesh_bind_band": "0",
# "amesh_bind_mac": "",
# "amesh_isRe": "0",
# "amesh_isReClient": "1",
# "amesh_papMac": "7C:10:C9:03:6D:90",
# "callback": "",
# "curRx": "        ",
# "curTx": "1",
# "defaultType": "0",
# "dpiDevice": "",
# "dpiType": "",
# "group": "",
# "internetMode": "block",
# "internetState": 0,
# "ip": "192.168.55.112",
# "ipMethod": "Manual",
# "isGN": "",
# "isGateway": "0",
# "isITunes": "0",
# "isLogin": "0",
# "isOnline": "1",
# "isPrinter": "0",
# "isWL": "1",
# "isWebServer": "0",
# "keeparp": "",
# "mac": "E0:98:06:C4:25:92",
# "macRepeat": "0",
# "name": "Espressif Inc ",
# "nickName": "Fr/Socket/Bookshelf",
# "opMode": "0",
# "os_type": 0,
# "qosLevel": "",
# "ROG": "0",
# "rssi": "-55",
# "ssid": "",
# "totalRx": "",
# "totalTx": "",
# "type": "0",
# "vendor": "Espressif Inc.",
# "vendorclass": "Espressif Inc.",
# "wtfast": "0",
# "wlConnectTime": "00:00:03"


@dataclass
class AsusClient:
    """Client class."""

    state: ConnectionState = ConnectionState.UNKNOWN

    description: Optional[AsusClientDescription] = None

    connection: Optional[AsusClientConnection] = None


@dataclass
class AsusClientConnection:
    """Connection class."""

    type: ConnectionType = ConnectionType.DISCONNECTED
    ip_address: Optional[str] = None
    ip_method: Optional[str] = None
    internet_state: Optional[bool] = None
    internet_mode: InternetMode = InternetMode.UNKNOWN
    node: Optional[str] = None
    online: bool = False


@dataclass
class AsusClientConnectionWlan(AsusClientConnection):
    """WLAN connection class."""

    guest: Optional[bool] = None
    guest_id: Optional[int] = None
    rssi: Optional[int] = None
    since: Optional[datetime] = None
    rx_speed: Optional[float] = None
    tx_speed: Optional[float] = None


@dataclass
class AsusClientDescription:
    """Client description class."""

    name: Optional[str] = None
    mac: Optional[str] = None

    vendor: Optional[str] = None


CLIENT_MAP = {
    "connected_since": [
        ("wlConnectTime", safe_time_from_delta),
    ],
    "connection_type": [
        ("connection_type"),
        ("isWL", safe_int),
    ],
    "guest": [
        ("guest"),
        ("isGN", safe_int),
    ],
    "internet_mode": [
        ("internetMode", get_internet_mode),
    ],
    "internet_state": [
        ("internetState", safe_bool),
    ],
    "ip": [
        ("ip"),
    ],
    "ip_method": [
        ("ipMethod"),
    ],
    "mac": [
        ("mac"),
    ],
    "name": [
        ("nickName"),
        ("name"),
        ("mac"),
    ],
    "node": [
        ("node"),
    ],
    "online": [
        ("online"),
        ("isOnline", safe_bool),
    ],
    "rssi": [
        ("rssi", safe_int),
    ],
    "rx_speed": [
        ("curRx", safe_float),
    ],
    "tx_speed": [
        ("curTx", safe_float),
    ],
}

CLIENT_MAP_DESCRIPTION = {
    "name": [
        ("nickName"),
        ("name"),
        ("mac"),
    ],
    "mac": [
        ("mac"),
    ],
    "vendor": [
        ("vendor", replace_vendor),
    ],
}

CLIENT_MAP_CONNECTION = {
    "type": [
        ("connection_type", get_connection_type),
        ("isWL", get_connection_type),
    ],
    "ip_address": [
        ("ip"),
    ],
    "ip_method": [
        ("ipMethod", read_ip_address_type),
    ],
    "internet_state": [
        ("internetState", safe_bool),
    ],
    "internet_mode": [
        ("internetMode", get_internet_mode),
    ],
    "node": [
        ("node"),
    ],
    "online": [
        ("online", safe_bool),
        ("isOnline", safe_bool),
    ],
}

CLIENT_MAP_CONNECTION_WLAN = {
    "guest_id": [
        ("guest"),
        ("isGN", safe_int),
    ],
    "rssi": [
        ("rssi", safe_int),
    ],
    "since": [
        ("wlConnectTime", safe_time_from_delta),
    ],
    "rx_speed": [
        ("curRx", safe_float),
    ],
    "tx_speed": [
        ("curTx", safe_float),
    ],
}


def process_client(
    data: dict[str, Any], history: Optional[AsusClient] = None
) -> AsusClient:
    """Process client data."""

    description = process_client_description(data)
    connection = process_client_connection(data)

    state = process_client_state(connection)

    # Clean disconnected client
    if state != ConnectionState.CONNECTED:
        connection = process_disconnected(connection)

    # Check history for the oscillating values
    if history is not None and history.connection is not None:
        connection = process_history(connection, history.connection)

    return AsusClient(state=state, description=description, connection=connection)


def process_client_description(data: dict[str, Any]) -> AsusClientDescription:
    """Process client description data."""

    description = AsusClientDescription()

    for key, value in CLIENT_MAP_DESCRIPTION.items():
        for pair in value:
            key_to_find, converter = safe_unpack_key(pair)
            item = data.get(key_to_find)
            if item and item != str():
                setattr(
                    description,
                    key,
                    converter(item) if converter else item,
                )
                break

    return description


def process_client_connection(data: dict[str, Any]) -> AsusClientConnection:
    """Process client connection data."""

    connection = AsusClientConnection()

    for key, value in CLIENT_MAP_CONNECTION.items():
        for pair in value:
            key_to_find, converter = safe_unpack_key(pair)
            item = data.get(key_to_find)
            if item and item != str():
                setattr(
                    connection,
                    key,
                    converter(item) if converter else item,
                )
                break

    if not connection.type in (ConnectionType.WIRED, ConnectionType.DISCONNECTED):
        connection = process_client_connection_wlan(data, connection)

    return connection


def process_client_connection_wlan(
    data: dict[str, Any], connection: AsusClientConnection
) -> AsusClientConnectionWlan:
    """Process WLAN client connection data."""

    connection = AsusClientConnectionWlan(**connection.__dict__)

    for key, value in CLIENT_MAP_CONNECTION_WLAN.items():
        for pair in value:
            key_to_find, converter = safe_unpack_key(pair)
            item = data.get(key_to_find)
            if item and item != str():
                setattr(
                    connection,
                    key,
                    converter(item) if converter else item,
                )
                break

    # Mark `guest` attribute if `guest_id` is non-zero
    connection.guest = connection.guest_id != 0 and connection.guest_id is not None

    return connection


def process_client_state(
    connection: AsusClientConnection,
) -> ConnectionState:
    """Process client state."""

    # If client has node attribute, it's connected
    if connection.node is not None:
        return ConnectionState.CONNECTED

    if connection.online is True:
        return ConnectionState.CONNECTED

    return ConnectionState.DISCONNECTED


def process_disconnected(connection: AsusClientConnection) -> AsusClientConnection:
    """Process disconnected client."""

    # We keep values set by the user:
    # - internet mode
    # - ip method
    # - ip address if ip method is static
    return AsusClientConnection(
        type=ConnectionType.DISCONNECTED,
        ip_address=connection.ip_address
        if connection.ip_method == IPAddressType.STATIC
        else None,
        ip_method=connection.ip_method,
        internet_mode=connection.internet_mode,
    )


def process_history(
    connection: AsusClientConnection, history: AsusClientConnection
) -> AsusClientConnection:
    """Process values from history to avoid oscillating values."""

    # Wireless connection
    if isinstance(connection, AsusClientConnectionWlan) and isinstance(
        history, AsusClientConnectionWlan
    ):
        # Fix the oscillating connection time ignore any value less than 10 seconds
        # This comes from the device itself and cannot be fixed from AsusRouter side
        if connection.since is not None and history.since is not None:
            diff = abs((history.since - connection.since).total_seconds())
            if diff < 10:
                connection.since = history.since

    return connection



"""State module."""

from __future__ import annotations

import importlib
import logging
from enum import Enum
from types import ModuleType
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.connection import ConnectionState
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.identity import AsusDevice
from asusrouter.modules.parental_control import AsusParentalControl
from asusrouter.modules.port_forwarding import AsusPortForwarding
from asusrouter.modules.system import AsusSystem
from asusrouter.modules.wlan import AsusWLAN

from .led import AsusLED
from .openvpn import AsusOVPNClient, AsusOVPNServer

_LOGGER = logging.getLogger(__name__)


class AsusStateNone(int, Enum):
    """Asus state none."""

    NONE = 0


# AsusState = Union[AsusLED, AsusOVPNClient, AsusOVPNServer, AsusStateNone]


class AsusState(Enum):
    """Asus state."""

    NONE = AsusStateNone
    CONNECTION = ConnectionState
    LED = AsusLED
    OPENVPN_CLIENT = AsusOVPNClient
    OPENVPN_SERVER = AsusOVPNServer
    PARENTAL_CONTROL = AsusParentalControl
    PORT_FORWARDING = AsusPortForwarding
    SYSTEM = AsusSystem
    WLAN = AsusWLAN


AsusStateMap: dict[AsusState, Optional[AsusData]] = {
    AsusState.NONE: None,
    AsusState.CONNECTION: None,
    AsusState.LED: AsusData.LED,
    AsusState.OPENVPN_CLIENT: AsusData.OPENVPN,
    AsusState.OPENVPN_SERVER: AsusData.OPENVPN,
    AsusState.PARENTAL_CONTROL: AsusData.PARENTAL_CONTROL,
    AsusState.PORT_FORWARDING: AsusData.PORT_FORWARDING,
    AsusState.SYSTEM: AsusData.SYSTEM,
    AsusState.WLAN: AsusData.WLAN,
}


def get_datatype(state: Optional[Any]) -> Optional[AsusData]:
    """Get the datatype."""

    asus_state = AsusState(type(state))

    if state is not None:
        return AsusStateMap.get(asus_state)

    return None


def _get_module_name(state: AsusState) -> Optional[str]:
    """Get the module name."""

    module_class = get_datatype(state)
    if module_class:
        return module_class.value

    return None


def _get_module(state: AsusState) -> Optional[ModuleType]:
    """Get the module."""

    # Module name
    module_name = _get_module_name(state)
    if not module_name:
        return None

    # Module path
    module_path = f"asusrouter.modules.{module_name}"

    try:
        # Import the module
        submodule = importlib.import_module(module_path)
        # Return the module
        return submodule
    except ModuleNotFoundError:
        _LOGGER.debug("No module found for state %s", AsusState)
        return None


def _has_method(module: ModuleType, method: str) -> bool:
    """Check if the module has the method."""

    return hasattr(module, method) and callable(getattr(module, method))


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusState,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the state."""

    # Get the module
    submodule = _get_module(state)

    # Process the data if module found
    if submodule and _has_method(submodule, "set_state"):
        return await submodule.set_state(callback, state, arguments, expect_modify)

    return False


def save_state(
    state: AsusState,
    library: dict[AsusData, AsusDataState],
) -> None:
    """Save the state."""

    # Get the correct data key
    datatype = get_datatype(state)
    if datatype is None:
        return

    # Save the state
    if datatype:
        library[datatype].update_state(state)


async def keep_state(
    callback: Callable[..., Awaitable[Any]],
    states: Optional[AsusState | list[AsusState]],
    identity: Optional[AsusDevice],
) -> None:
    """Keep the state."""

    if states is None:
        return

    # Make sure the state is a list
    if not isinstance(states, list):
        states = [states]

    # Process each state
    for state in states:
        # Get the module
        submodule = _get_module(state)

        # Process the data if module found
        if submodule and _has_method(submodule, "keep_state"):
            await submodule.keep_state(callback, state, identity)



"""Firmware module."""

from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import Optional

from asusrouter.tools.converters import clean_string

_LOGGER = logging.getLogger(__name__)


@dataclass
class Firmware:
    """Firmware class.

    This class contains information about the firmware of a device."""

    major: str = "3.0.0.4"
    minor: Optional[int] = None
    build: Optional[int] = None
    revision: Optional[int | str] = None

    def __str__(self) -> str:
        """Return the firmware version as a string."""

        return f"{self.major}.{self.minor}.{self.build}_{self.revision}"

    def __lt__(self, other: Firmware | None) -> bool:
        """Compare two firmware versions."""

        result = False

        if other:
            if self.major and other.major:
                # Check if the first character of either major version is the same
                if self.major[0] == other.major[0]:
                    # Split the major versions into lists of integers
                    major1 = [int(x) for x in self.major.split(".")]
                    major2 = [int(x) for x in other.major.split(".")]
                    # Compare the major versions
                    if major1 < major2:
                        return True
                    # Proceed only if major is the same for both
                    if major1 == major2:
                        # Compare the minor versions
                        if self.minor and other.minor and self.minor < other.minor:
                            return True
                        # Compare the build versions
                        if self.build and other.build and self.build < other.build:
                            return True
                        # Compare the revision versions
                        if (
                            isinstance(self.revision, int)
                            and isinstance(other.revision, int)
                            and self.revision < other.revision
                        ):
                            return True

        return result

    # Define more-than
    def __gt__(self, other: Firmware | None) -> bool:
        """Compare two firmware versions."""

        # Invert the statement of less-than
        return not self.__lt__(other)


def read_fw_string(content: Optional[str]) -> Optional[Firmware]:
    """Firmware string parser"""

    # Clean and check the string
    content = clean_string(content)
    if not content:
        return None

    # Search pattern
    pattern = (
        r"^(?P<major>[39].?0.?0.?[46])?[_.]?"
        r"(?P<minor>[0-9]{3})[_.]?"
        r"(?P<build>[0-9]+)[_.-]?"
        r"(?P<revision>[a-zA-Z0-9-_]+)?$"
    )
    match = re.match(pattern, content)
    if not match:
        _LOGGER.warning(
            "Firmware version cannot be parsed. Please report this. The original FW string is: %s",
            content,
        )
        return Firmware()

    major = match.group("major") or "3.0.0.4"
    major = (
        major[0] + "." + major[1] + "." + major[2] + "." + major[3]
        if major and not "." in major and len(major) == 4
        else major
    )

    minor = int(match.group("minor"))

    build = int(match.group("build"))

    revision = match.group("revision")
    revision = (
        int(revision)
        if revision and revision.isdigit()
        else revision
        if revision
        else 0
    )

    firmwawe = Firmware(
        major=major,
        minor=minor,
        build=build,
        revision=revision,
    )

    return firmwawe



"""Parental control module."""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

from asusrouter.tools.converters import safe_int, safe_return

KEY_PARENTAL_CONTROL_MAC = "MULTIFILTER_MAC"
KEY_PARENTAL_CONTROL_NAME = "MULTIFILTER_DEVICENAME"
KEY_PARENTAL_CONTROL_STATE = "MULTIFILTER_ALL"
KEY_PARENTAL_CONTROL_TIMEMAP = "MULTIFILTER_MACFILTER_DAYTIME_V2"
KEY_PARENTAL_CONTROL_TYPE = "MULTIFILTER_ENABLE"

MAP_PARENTAL_CONTROL_ITEM = [
    (KEY_PARENTAL_CONTROL_MAC, "mac", safe_return),
    (KEY_PARENTAL_CONTROL_NAME, "name", safe_return),
    (KEY_PARENTAL_CONTROL_TIMEMAP, "timemap", safe_return),
    (KEY_PARENTAL_CONTROL_TYPE, "type", safe_int),
]

MAP_PARENTAL_CONTROL_TYPE = {
    0: "disable",
    1: "time",
    2: "block",
}

DEFAULT_PARENTAL_CONTROL_TIMEMAP = "W03E21000700<W04122000800"


@dataclass(frozen=True)
class ParentalControlRule:
    """Parental control rule class."""

    mac: Optional[str] = None
    name: Optional[str] = None
    type: Optional[str] = None
    timemap: Optional[str] = DEFAULT_PARENTAL_CONTROL_TIMEMAP


class AsusParentalControl(IntEnum):
    """Asus parental control state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusParentalControl,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the parental control state."""

    # Check if arguments are available
    if not arguments:
        arguments = {}

    arguments[KEY_PARENTAL_CONTROL_STATE] = 1 if state == AsusParentalControl.ON else 0

    # Get the correct service call
    service = "restart_firewall"

    # Call the service
    return await callback(
        service, arguments=arguments, apply=True, expect_modify=expect_modify
    )



"""Ethernet ports endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.modules.ports import PORT_SPEED, PortSpeed, PortType
from asusrouter.tools.converters import safe_int
from asusrouter.tools.readers import read_json_content


def read(content: str) -> dict[str, Any]:
    """Read ethernet ports data."""

    # Read the json content
    ethernet_ports: dict[str, Any] = read_json_content(
        content.replace("get_wan_lan_status = ", "").replace(";", "")
    )

    return ethernet_ports


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process ethernet ports data."""

    # Ports info
    ports = {
        PortType.LAN: {},
        PortType.WAN: {},
    }
    if "portSpeed" in data:
        data = data["portSpeed"]
        for port, value in data.items():
            port_type = PortType(port[0:3].lower())
            port_id = safe_int(port[3:])
            link_rate = PORT_SPEED.get(value)
            ports[port_type][port_id] = {
                "state": link_rate != PortSpeed.LINK_DOWN,
                "link_rate": link_rate,
            }

    return {
        AsusData.PORTS: ports,
    }



"""Data transform module."""

from __future__ import annotations

from typing import Any, Optional

from asusrouter.modules.client import process_client
from asusrouter.modules.data import AsusDataState
from asusrouter.tools.readers import readable_mac


def transform_network(
    data: dict[str, Any],
    services: Optional[list[str]],
    history: Optional[AsusDataState],
) -> dict[str, Any]:
    """Transform network data."""

    # Check if the device has dualwan support
    if not services:
        return data
    if not "dualwan" in services:
        return data

    network = data.copy()
    # Add speed if not available - fix first empty value round
    for interface in network:
        for speed in ("rx_speed", "tx_speed"):
            if not speed in network[interface]:
                network[interface][speed] = 0.0

    # Add usb network if not available
    if not "usb" in network:
        # Check history
        usb_history = history.data.get("usb") if history and history.data else None
        # Revert to history if available
        if usb_history:
            network["usb"] = usb_history
            network["usb"]["rx_speed"] = 0.0
            network["usb"]["tx_speed"] = 0.0
        else:
            network["usb"] = {
                "rx": 0,
                "tx": 0,
                "rx_speed": 0.0,
                "tx_speed": 0.0,
            }
    return network


def transform_clients(
    data: dict[str, Any],
    history: Optional[AsusDataState],
) -> dict[str, Any]:
    """Transform clients data."""

    clients = {}
    for mac, client in data.items():
        if readable_mac(mac):
            # Check client history
            client_history = history.data.get(mac) if history and history.data else None
            # Process the client
            clients[mac] = process_client(client, client_history)

    return clients



"""Constants for hook endpoint module."""

# Network

from asusrouter.modules.ip_address import read_dns_ip_address, read_ip_address_type
from asusrouter.tools.converters import safe_bool, safe_int

MAP_NETWORK: dict[str, str] = {
    "INTERNET": "wan",
    "INTERNET1": "usb",
    "WIRED": "wired",
    "BRIDGE": "bridge",
    "WIRELESS0": "2ghz",
    "WIRELESS1": "5ghz",
    "WIRELESS2": "5ghz2",
    "WIRELESS3": "6ghz",
    "LACP1": "lacp1",
    "LACP2": "lacp2",
}

MAP_WAN = (
    ("wanstate", "state"),
    ("wansbstate", "bstate"),
    ("wanauxstate", "aux"),
    ("autodet_state"),
    ("autodet_auxstate"),
    ("wanlink_status", "status", safe_bool),
    ("wanlink_type", "ip_type", read_ip_address_type),
    ("wanlink_ipaddr", "ip_address"),
    ("wanlink_netmask", "mask"),
    ("wanlink_gateway", "gateway"),
    ("wanlink_dns", "dns", read_dns_ip_address),
    ("wanlink_lease", "lease", safe_int),
    ("wanlink_expires", "expires", safe_int),
    ("is_private_subnet", "private_subnet", safe_int),
    ("wanlink_xtype", "xtype", read_ip_address_type),
    ("wanlink_xipaddr", "xip"),
    ("wanlink_xnetmask", "xmask"),
    ("wanlink_xgateway", "xgateway"),
    ("wanlink_xdns", "xdns", read_dns_ip_address),
    ("wanlink_xlease", "xlease", safe_int),
    ("wanlink_xexpires", "xexpires", safe_int),
)



"""Service module."""

from __future__ import annotations

import logging
from typing import Any, Awaitable, Callable, Optional

from asusrouter.error import AsusRouterServiceError
from asusrouter.tools.converters import safe_bool

_LOGGER = logging.getLogger(__name__)


async def async_call_service(
    callback: Callable[..., Awaitable[dict[str, Any]]],
    service: str,
    arguments: Optional[dict[str, Any]] = None,
    apply: bool = False,
    expect_modify: bool = True,
) -> bool:
    """Call a service."""

    # Generate commands
    commands = {
        "rc_service": service,
    }

    # Check arguments
    if not arguments:
        arguments = {}

    # Add apply command if requested
    if apply:
        arguments["action_mode"] = "apply"

    # Add arguments to the commands
    commands.update(arguments)

    # Send the commands
    try:
        result = await callback(commands)
    except Exception as ex:  # pylint: disable=broad-except
        raise ex

    run_service = result.get("run_service", None)
    if run_service != service:
        raise AsusRouterServiceError(f"Service not run. Raw result: {result}")

    _LOGGER.debug(
        "Service `%s` run with arguments `%s`. Result: `%s`", service, arguments, result
    )

    if expect_modify:
        return safe_bool(result.get("modify")) or False

    return True



"""Port forwarding module."""

from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

KEY_PORT_FORWARDING_LIST = "vts_rulelist"
KEY_PORT_FORWARDING_STATE = "vts_enable_x"


@dataclass(frozen=True)
class PortForwardingRule:
    """Port forwarding class"""

    name: str = str()
    ip_address: str = str()
    port: str = str()
    protocol: str = str()
    ip_external: str = str()
    port_external: str = str()


class AsusPortForwarding(IntEnum):
    """Asus port forwarding state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusPortForwarding,
    arguments: Optional[dict[str, Any]] = None,
    expect_modify: bool = False,
) -> bool:
    """Set the parental control state."""

    # Check if arguments are available
    if not arguments:
        arguments = {}

    arguments[KEY_PORT_FORWARDING_STATE] = 1 if state == AsusPortForwarding.ON else 0

    # Get the correct service call
    service = "restart_firewall"

    # Call the service
    return await callback(
        service, arguments=arguments, apply=True, expect_modify=expect_modify
    )



"""Port module for AsusRouter.

This module is for physical ports on the router, e.g. LAN, WAN, USB, etc."""

from enum import Enum

from asusrouter.tools.converters import safe_bool, safe_int


class PortSpeed(int, Enum):
    """Port speed class."""

    LINK_DOWN = 0
    LINK_100 = 100
    LINK_1000 = 1000
    LINK_2500 = 2500
    LINK_10000 = 10000


PORT_SPEED = {
    "X": PortSpeed.LINK_DOWN,
    "M": PortSpeed.LINK_100,
    "G": PortSpeed.LINK_1000,
    "Q": PortSpeed.LINK_2500,
}


class PortType(str, Enum):
    """Port type class."""

    LAN = "lan"
    USB = "usb"
    WAN = "wan"


PORT_TYPE = {
    "L": PortType.LAN,
    "U": PortType.USB,
    "W": PortType.WAN,
}

PORT_VALUE_MAP = [
    ("is_on", "state", safe_bool),
    ("max_rate", "max_rate", safe_int),
    ("link_rate", "link_rate", safe_int),
]



"""Converters module

This module has methods to convert data between different formats 
without complicated logic. In case data cannot be converted, 
`None` is returned and no exception is raised.

If data conversion requires complicated logic, 
it should be in the Readers module"""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Optional

from dateutil.parser import parse as dtparse


def as_dict(pyobj):
    """Return generator object as dictionary."""

    return dict(pyobj)


def clean_string(content: Optional[str]) -> Optional[str]:
    """Get a clean string or return None if it is empty."""

    # Not a string
    if not content or not isinstance(content, str):
        return None

    content = content.strip()
    # Empty string
    if not content:
        return None

    return content


def flatten_dict(obj: Any, keystring: str = "", delimiter: str = "_"):
    """Flatten dictionary."""

    if isinstance(obj, dict):
        keystring = keystring + delimiter if keystring else keystring
        for key in obj:
            yield from flatten_dict(obj[key], keystring + str(key))
    else:
        yield keystring, obj


def list_from_dict(raw: dict[str, Any]) -> list[str]:
    """Return dictionary keys as list."""

    return list(raw.keys())


def nvram_get(content: Optional[list[str] | str]) -> Optional[list[tuple[str, ...]]]:
    """Convert values to NVRAM request."""

    if not content:
        return None

    if isinstance(content, str):
        content = [content]

    return [("nvram_get", value) for value in content]


def safe_bool(content: Optional[str | int | float | bool]) -> Optional[bool]:
    """Read the content as boolean or return None."""

    result = None

    if content:
        if isinstance(content, bool):
            result = content
        elif isinstance(content, (int, float)):
            result = content != 0
        else:
            content = clean_string(content)
            if content:
                content = content.lower()
                if content in ("true", "allow", "1", "on", "enabled"):
                    result = True
                elif content in ("false", "block", "0", "off", "disabled"):
                    result = False

    return result


def safe_datetime(content: Optional[str]) -> Optional[datetime]:
    """Read the content as datetime or return None."""

    content = clean_string(content)
    if not content:
        return None

    try:
        return dtparse(content)
    except Exception:  # pylint: disable=broad-except
        return None


def safe_exists(content: Optional[str]) -> bool:
    """Read the content as boolean or return None."""

    content = clean_string(content)
    if not content:
        return False

    return True


def safe_float(
    content: Optional[str | int | float], default: Optional[float] = None
) -> Optional[float]:
    """Read the content as float or return None."""

    if isinstance(content, (int, float)):
        return float(content)

    content = clean_string(content)
    if not content:
        return default

    try:
        return float(content)
    except ValueError:
        return default


def safe_int(
    content: Optional[str | int], default: Optional[int] = None, base: int = 10
) -> Optional[int]:
    """Read the content as int or return the default value (None if not specified)."""

    if isinstance(content, int):
        return content

    content = clean_string(content)
    if not content:
        return default if isinstance(default, int) else None

    try:
        return int(content, base=base)
    except ValueError:
        return default if isinstance(default, int) else None


def safe_list(content: Any) -> list[Any]:
    """Read any content as a list."""

    if isinstance(content, list):
        return content

    if content is None:
        return []

    return [content]


def safe_list_from_string(content: Optional[str], delimiter: str = " ") -> list[str]:
    """Read the content as list or return empty list."""

    content = clean_string(content)
    if not content:
        return []

    return content.split(delimiter)


def safe_none_or_str(content: Optional[str]) -> Optional[str]:
    """Read the content as string or return None if it is empty."""

    if content is None or not isinstance(content, str):
        return None

    content = content.strip()
    if content == str():
        return None

    return content


def safe_return(content: Any) -> Any:
    """Return the content."""

    if isinstance(content, str):
        return clean_string(content)

    return content


def safe_speed(
    current: (int | float),
    previous: (int | float),
    time_delta: (int | float) | None = None,
) -> float:
    """Calculate speed.

    Allows calculation only of positive speed, otherwise returns 0.0."""

    if time_delta is None or time_delta == 0.0:
        return 0.0

    diff = current - previous if current > previous else 0.0

    return diff / time_delta


def safe_time_from_delta(content: str) -> datetime:
    """Transform time delta to the date in the past."""

    return datetime.utcnow().replace(
        microsecond=0, tzinfo=timezone.utc
    ) - safe_timedelta_long(content)


def safe_timedelta_long(content: Optional[str]) -> timedelta:
    """Transform connection timedelta of the device to a proper
    datetime object when the device was connected"""

    content = clean_string(content)
    if not content:
        return timedelta()

    part = content.split(":")
    try:
        return timedelta(
            hours=int(part[-3]), minutes=int(part[-2]), seconds=int(part[-1])
        )
    except ValueError:
        return timedelta()


def safe_unpack_key(
    content: tuple[str, Optional[Callable[..., Any]]] | str
) -> tuple[str, Optional[Callable[..., Any]]]:
    """Method to unpack key/method tuple
    even if some values are missing."""

    if isinstance(content, tuple):
        # All 2 values are present
        if len(content) == 2:
            return content

    # No method selected
    return (content, None)


def safe_unpack_keys(
    content: tuple[str, str, Callable[..., Any]] | tuple[str, str] | str
) -> tuple[Any, ...]:
    """Method to unpack key/key_to_use/method tuple
    even if some values are missing."""

    if isinstance(content, tuple):
        # All 3 values are present
        if len(content) == 3:
            return content

        # No method selected
        if len(content) == 2:
            return content + (None,)

    # No method and key_to_use selected
    # We need to replace key_to_use with key
    content = (content, content)
    return content + (None,)


def safe_usage(used: int | float, total: int | float) -> float:
    """Calculate usage in percents.

    Allows calculation only of positive usage, otherwise returns 0.0."""

    if total == 0:
        return 0.0

    usage = round(used / total * 100, 2)

    # Don't allow negative usage
    if usage < 0:
        return 0.0

    return usage


def safe_usage_historic(
    used: int | float,
    total: int | float,
    prev_used: int | float,
    prev_total: int | float,
) -> float:
    """Calculate usage in percents for difference between current and previous values

    This method is just an interface to calculate usage using `usage` method"""

    return safe_usage(used - prev_used, total - prev_total)



"""Logger module for AsusRouter.

This module is needed to add a TRACE logging level to the logger,
which provides more information than DEBUG. In general, this information
is not needed for general use, but can help in tracing issues.

This module also adds an UNSAFE logging level, which can provide sensitive
information, including passwords. Not recommended for general use."""

from __future__ import annotations

import logging

TRACE = 5
UNSAFE = 3

logging.addLevelName(TRACE, "TRACE")
logging.addLevelName(UNSAFE, "UNSAFE")


def trace(self, message, *args, **kws):
    """Trace logging level."""

    if self.isEnabledFor(TRACE):
        self._log(TRACE, message, args, **kws)  # pylint: disable=protected-access


def unsafe(self, message, *args, **kws):
    """Unsafe logging level."""

    if self.isEnabledFor(UNSAFE):
        self._log(UNSAFE, message, args, **kws)  # pylint: disable=protected-access


logging.Logger.trace = trace
logging.Logger.unsafe = unsafe



"""Tools module"""


def get_enum_key_by_value(enum, value, default=None):
    """Get the enum key by value"""

    for key, enum_value in enum.__members__.items():
        if enum_value.value == value:
            return enum[key]

    return default
